import json
import os
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Dict, Optional, Tuple

import agate
import requests

import dbt.exceptions
from dbt.adapters.base import BaseConnectionManager, Credentials
from dbt.adapters.factory import get_adapter_class_by_name, load_plugin
from dbt.contracts.connection import AdapterResponse
from dbt.logger import GLOBAL_LOGGER as logger


@dataclass
class InferAdapterResponse(AdapterResponse):
    something: Optional[int] = None


@dataclass
class InferCredentials(Credentials):

    apikey: str
    data_config: Dict[str, Any]

    _ALIASES = {"url": "database", "username": "schema"}

    def __init__(self, database: str, schema: str, apikey: str, data_config: Dict[str, Any]):
        self.url = os.getenv("INFER_URL", database)
        self.username = os.getenv("INFER_USER", schema)
        self.apikey = os.getenv("INFER_KEY", apikey)
        self.data_config = data_config
        # setting up the adapter class before we use it
        from . import Plugin

        Plugin.dependencies = [data_config["type"]]
        from .impl import InferAdapter

        InferAdapter.SourceAdapter = InferConnectionManager.get_source_module(data_config)[0]
        InferAdapter.Relation = InferAdapter.SourceAdapter.Relation
        InferAdapter.Column = InferAdapter.SourceAdapter.Column

    def __getattr__(self, name):
        if name == "adapter_credentials":
            _, cls = InferConnectionManager.get_source_module(self.data_config)
            data = cls.translate_aliases(self.data_config)
            cls.validate(data)
            self.adapter_credentials = cls.from_dict(data)
            return self.adapter_credentials
        return getattr(self.adapter_credentials, name)

    def __getattribute__(self, item):
        if item == "database":
            return self.data_config.database
        elif item == "schema":
            return self.data_config.schema
        elif item == "project":
            return self.data_config.project.strip("\"'")
        return object.__getattribute__(self, item)

    @property
    def type(self):
        return "infer"

    @property
    def unique_field(self):
        return self.database

    def _connection_keys(self):
        return ("database", "schema")


class InferSession:
    def __init__(self, credentials):
        session = requests.Session()
        session.headers.update(
            {
                "Authorization": f'Token token="{credentials.apikey}",'
                f"email={credentials.username}",
                "Content-Type": "application/json",
            }
        )
        url = f"{credentials.url}/api/v1"
        r = session.get(f"{url}/users/me")
        if r.status_code != 200:
            raise RuntimeError(f"Failed to connect to Infer server {url} end point 'users/me'")
        self.__baseurl = credentials.url
        self.__url = url
        self.__session = session

    def parse(self, sql):
        r = self.__session.post(f"{self.__url}/parse", json={"q": sql})
        if r.status_code != 200:
            raise RuntimeError(
                f"Failed to connect to Infer server {self.__url} end "
                f"point 'parse' got return code {r.status_code}"
            )
        return r.json()["result"]

    def single_dataset_run(self, dataset_id, name, query):
        r = self.__session.post(
            f"{self.__url}/datasets/{dataset_id}/results",
            json={"result": {"name": name, "query": query}},
        )
        if r.status_code != 200:
            raise RuntimeError(
                f"Failed to connect to Infer server {self.__url} "
                f"to run query={query} on dataset_id={dataset_id} "
                f"got return code {r.status_code}"
            )
        return r.json()["id"]

    def dbt_run(self, name, query, datasets):
        r = self.__session.post(
            f"{self.__url}/dbt_runs",
            json={
                "dbt_run": {
                    "name": name,
                    "description": name,
                    "query": query,
                    "datasets": datasets,
                }
            },
        )
        if r.status_code != 200:
            raise RuntimeError(
                f"Failed to run `dbt_run` on {self.__url} "
                f"with query={query} got return code {r.status_code}"
            )
        return r.json()["id"]

    def get_dbt_result(self, result_id):
        r = self.__session.get(f"{self.__url}/dbt_runs/{result_id}")
        if r.status_code != 200:
            raise RuntimeError(
                f"Failed to connect to Infer server {self.__url} "
                f"to retrieve result_id={result_id} "
                f"got return code {r.status_code}"
            )
        r_json = r.json()
        r_status = r_json["status"]
        rtn_obj = None
        if r_status == "COMPLETED":
            url = f"{self.__baseurl}{r_json['output_url']}"
            r = self.__session.get(url)
            if r.status_code != 200:
                raise RuntimeError(
                    f"Failed to connect to retrieve result {url} "
                    f"got return code {r.status_code}"
                )
            rtn_obj = r.content
        elif r_status == "ERROR":
            rtn_obj = r_json["raw_output"]
        return r_status, rtn_obj

    def get_result(self, result_id):
        r = self.__session.get(f"{self.__url}/results/{result_id}")
        if r.status_code != 200:
            raise RuntimeError(
                f"Failed to connect to Infer server {self.__url} "
                f"to retrieve result_id={result_id} "
                f"got return code {r.status_code}"
            )
        r_json = r.json()
        r_status = r_json["status"]
        rtn_obj = None
        if r_status == "COMPLETED":
            url = f"{self.__baseurl}{r_json['result_url']}"
            r = self.__session.get(url)
            if r.status_code != 200:
                raise RuntimeError(
                    f"Failed to connect to retrieve result {url} "
                    f"got return code {r.status_code}"
                )
            rtn_obj = r.content
        return r_status, rtn_obj

    def delete_dataset(self, dataset_id):
        r = self.__session.delete(f"{self.__url}/datasets/{dataset_id}")
        if r.status_code not in [200, 204]:
            raise RuntimeError(
                f"Failed to connect to Infer server {self.__url} "
                f"end point 'datasets/{dataset_id}' got "
                f"return code {r.status_code}"
            )

    def create_dataset(self, name, query, encode_data):
        data = {
            "dataset": {
                "name": name,
                "description": query,
                "source_file": {"data": encode_data},
            }
        }
        r = self.__session.post(f"{self.__url}/datasets", json=data)
        if r.status_code != 200:
            raise RuntimeError(
                f"Failed to connect to Infer server {self.__url} "
                f"end point 'datasets' got return code {r.status_code}"
            )
        return r.json()["id"]

    def close(self):
        self.__session.close()


class InferConnectionManager(BaseConnectionManager):
    TYPE = "infer"

    @contextmanager
    def exception_handler(self, sql: str):
        try:
            yield
        except Exception as exc:
            logger.debug("myadapter error: {}".format(str(exc)))
            raise dbt.exceptions.DatabaseException(str(exc))
        except Exception as exc:
            logger.debug("Error running SQL: {}".format(sql))
            logger.debug("Rolling back transaction.")
            raise dbt.exceptions.RuntimeException(str(exc))

    @classmethod
    def get_source_module(cls, source):
        source_type = source["type"]
        credentials = load_plugin(source_type)
        return get_adapter_class_by_name(source_type), credentials

    @classmethod
    def open(cls, connection):
        if connection.state == "open":
            logger.debug("Connection is already open, skipping open.")
            return connection

        credentials = connection.credentials
        try:
            connection.state = "open"
            data_adapter, _ = cls.get_source_module(connection.credentials.data_config)
            connection.handle = {
                "session": InferSession(credentials),
                "data_adapter": data_adapter,
            }
        except Exception as e:
            raise e
        return connection

    @classmethod
    def close(cls, connection):
        if connection.state == "open":
            connection.handle["session"].close()
            connection.state = "closed"
        return connection

    def begin(self):
        pass

    def commit(self):
        pass

    def cancel_open(self) -> None:
        pass

    def execute(
        self, sql, auto_begin=False, fetch=None
    ) -> Tuple[InferAdapterResponse, agate.Table]:
        raise dbt.exceptions.NotImplementedException(
            "`execute` not implemented on InferConnectionManager"
        )

    @classmethod
    def get_response(cls, cursor):
        # we do not support rich metadata
        pass

    def cancel(self, connection):
        # we do not support cancelling a request
        pass



import base64
import io
import os
import uuid
from time import sleep
from typing import Any, Dict, List, Optional, Tuple

import agate

from dbt.adapters.base import BaseAdapter
from dbt.adapters.base import Column as BaseColumn
from dbt.adapters.base.meta import available
from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.infer import InferConnectionManager
from dbt.clients import agate_helper
from dbt.contracts.connection import AdapterResponse, Connection
from dbt.contracts.results import RunStatus
from dbt.events import AdapterLogger
from dbt.exceptions import RuntimeException
from dbt.task.seed import SeedTask

logger = AdapterLogger("Infer")


class InferAdapter(BaseAdapter):
    SourceAdapter = None
    ConnectionManager = InferConnectionManager
    Relation = None
    Column = None

    def __init__(self, config):
        super().__init__(config)
        self.config = config
        self.data_adapter = None
        self.create_view_mode = False

    @classmethod
    def is_cancelable(cls) -> bool:
        return False

    @available.parse(lambda *a, **k: (None, None))
    def add_query(
        self,
        sql: str,
        auto_begin: bool = True,
        bindings: Optional[Any] = None,
        abridge_sql_log: bool = False,
    ) -> Tuple[Connection, Any]:
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.add_query(sql, auto_begin, bindings, abridge_sql_log)

    @classmethod
    def date_function(cls):
        return cls.SourceAdapter.date_function()

    @available.parse(lambda *a, **k: ("", agate_helper.empty_table()))
    def execute(
        self, sql: str, auto_begin: bool = False, fetch: bool = False
    ) -> Tuple[AdapterResponse, agate.Table]:
        adapter = self.get_data_adapter()
        thread_connection = self.connections.get_if_exists()
        if not thread_connection:
            # we assume that we are in a nested execution
            with adapter.connection_named("master"):
                return adapter.execute(sql, auto_begin, fetch)
        data_source = thread_connection.handle
        session = data_source["session"]
        parsed_sql = session.parse(sql)
        if parsed_sql.get("errors", None):
            error = parsed_sql["errors"][0]
            logger.info(f"Failed to parse SQL as SQL-inf: ({error['type']}) {error['value']}")
            logger.info(f"Will try to process with {adapter.__class__.__name__}.")
        if not parsed_sql["infer_commands"]:
            with adapter.connection_named("master"):
                return adapter.execute(sql, auto_begin, fetch)

        if self.create_view_mode:
            raise RuntimeException("SQL-inf commands can only be used with TABLE materializations")

        datasets = []
        with adapter.connection_named("load_queries"):
            for query in parsed_sql["load_queries"]:
                result = adapter.execute(query, False, True)
                dataset_name = "tmp_" + str(uuid.uuid4()).replace("-", "")
                fp = io.StringIO()
                result[1].to_csv(fp)
                encoded_fp = base64.b64encode(fp.getvalue().encode())
                datasets.append(
                    {
                        "base64": encoded_fp.decode(),
                        "filename": dataset_name,
                        "table_query": query,
                    }
                )
                fp.close()

        result_id = session.dbt_run(name="dbt_run", query=sql, datasets=datasets)

        keep_running = True
        result = {}
        result_status = "STARTED"
        while keep_running:
            result_status, result = session.get_dbt_result(result_id)
            keep_running = result_status in ["STARTED", "RUNNING"]
            sleep(3)
        if result_status == "ERROR":
            raise RuntimeException(
                f"Failed to run SQL-inf command: "
                f"({result.get('type', 'Unknown Internal Error')}) {result.get('error', '')}"
            )
        if not result:
            raise RuntimeException(f"Failed to get result for SQL-inf command sql={sql}")

        temp_table_name = "tmp_infer_" + str(uuid.uuid4()).replace("-", "")
        output_file_path = f"seeds/{temp_table_name}.csv"
        with open(output_file_path, "w+b") as fp:
            fp.write(result)

        class SeedArgs:
            state = None
            single_threaded = True
            selector_name = None
            select = [temp_table_name]
            exclude = None
            show = None

        task = SeedTask(SeedArgs(), adapter.config)
        r = task.run()
        if r.results[0].status == RunStatus.Error:
            raise RuntimeException(r.results[0].message)

        os.remove(output_file_path)

        with adapter.connection_named("upload_infer_results"):
            full_temp_table_name = f"{adapter.config.credentials.schema}.{temp_table_name}"
            outer_sql = parsed_sql["outer"].replace(
                "__INNER_SELECT__", f"SELECT * FROM {full_temp_table_name}"
            )
            outer_result = adapter.execute(outer_sql, False, True)

            database = adapter.config.credentials.database
            schema = adapter.config.credentials.schema
            relation = adapter.Relation.create(
                database=database,
                schema=schema,
                identifier=temp_table_name,
                type="table",
                quote_policy=adapter.config.quoting,
            )

        adapter.drop_relation(relation)

        return outer_result

    def get_data_adapter(self):
        if not self.data_adapter:
            data_source = self.connections.get_thread_connection().handle
            self.data_adapter = data_source["data_adapter"](self.config)
        return self.data_adapter

    @available.parse(lambda *a, **k: {})
    def set_create_view_mode(self, view_mode):
        self.create_view_mode = view_mode

    @available.parse(lambda *a, **k: {})
    def adapter_macro(self, macro, macro_dict):
        data_adapter = self.get_data_adapter()
        return data_adapter.execute_macro(
            f"{data_adapter.type()}__{macro}",
            kwargs=macro_dict,
            manifest=data_adapter._macro_manifest,
        )

    @available.parse(lambda *a, **k: {})
    def get_view_options(self, config, node):
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.get_common_options(config, node)

    def list_relations_without_caching(self, schema_relation: BaseRelation) -> List[BaseRelation]:
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.list_relations_without_caching(schema_relation)

    def __getattribute__(self, item):
        if item == "_available_":
            adapter = self.get_data_adapter()
            return frozenset().union(adapter._available_, object.__getattribute__(self, item))
        if item == "database":
            return self.data_config.database
        elif item == "schema":
            return self.data_config.schema
        elif item == "project":
            return self.data_config.project.strip("\"'")
        return object.__getattribute__(self, item)

    def __getattr__(self, name):
        return getattr(self.get_data_adapter(), name)

    @available.parse(lambda *a, **k: True)
    def is_replaceable(self, relation, conf_partition, conf_cluster) -> bool:
        adapter = self.get_data_adapter()
        return adapter.is_replaceable(relation, conf_partition, conf_cluster)

    @available.parse(lambda *a, **k: {})
    def get_table_options(
        self, config: Dict[str, Any], node: Dict[str, Any], temporary: bool
    ) -> Dict[str, Any]:
        adapter = self.get_data_adapter()
        return adapter.get_table_options(config, node, temporary)

    @available
    def parse_partition_by(self, raw_partition_by: Any) -> Any:
        adapter = self.get_data_adapter()
        return adapter.parse_partition_by(raw_partition_by)

    @classmethod
    def convert_text_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return cls.SourceAdapter.convert_text_type(agate_table, col_idx)

    @classmethod
    def convert_number_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return cls.SourceAdapter.convert_number_type(agate_table, col_idx)

    @classmethod
    def convert_boolean_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return cls.SourceAdapter.convert_boolean_type(agate_table, col_idx)

    @classmethod
    def convert_datetime_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return cls.SourceAdapter.convert_datetime_type(agate_table, col_idx)

    @classmethod
    def convert_date_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return cls.SourceAdapter.convert_date_type(agate_table, col_idx)

    @classmethod
    def convert_time_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return cls.SourceAdapter.convert_time_type(agate_table, col_idx)

    @available.parse_none
    def create_schema(self, relation: BaseRelation):
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.create_schema(relation)

    @available.parse_none
    def drop_schema(self, relation: BaseRelation):
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.drop_schema(relation)

    @available
    @classmethod
    def quote(cls, identifier: str) -> str:
        return cls.SourceAdapter.quote(identifier)

    def quote(self, identifier: str) -> str:
        return self.get_data_adapter().quote(identifier)

    def expand_column_types(self, goal: BaseRelation, current: BaseRelation) -> None:
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.expand_column_types(goal, current)

    @available.parse_none
    def drop_relation(self, relation: BaseRelation) -> None:
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.drop_relation(relation)

    @available.parse_none
    def truncate_relation(self, relation: BaseRelation) -> None:
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.truncate_relation(relation)

    @available.parse_none
    def rename_relation(self, from_relation: BaseRelation, to_relation: BaseRelation) -> None:
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.rename_relation(from_relation, to_relation)

    @available.parse_list
    def get_columns_in_relation(self, relation: BaseRelation) -> List[BaseColumn]:
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.get_columns_in_relation(relation)

    def list_schemas(self, database: str) -> List[str]:
        adapter = self.get_data_adapter()
        with adapter.connection_named("master"):
            return adapter.list_schemas(database.strip("\"'"))



from dbt.adapters.base import AdapterPlugin
from dbt.adapters.infer.connections import InferConnectionManager  # noqa
from dbt.adapters.infer.connections import InferCredentials
from dbt.adapters.infer.impl import InferAdapter
from dbt.include import infer

Plugin = AdapterPlugin(
    adapter=InferAdapter,
    credentials=InferCredentials,
    include_path=infer.PACKAGE_PATH,
    dependencies=[],
)



import os

from setuptools import find_namespace_packages, setup

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md")) as f:
    long_description = f.read()

setup(
    name="dbt-infer",
    version="1.2.1",
    description="The Infer adapter plugin for dbt",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Infer",
    author_email="support@inferlabs.io",
    url="https://github.com/inferlabs/dbt-infer",
    packages=find_namespace_packages(include=["dbt", "dbt.*"]),
    include_package_data=True,
    install_requires=["dbt-core>=1.2.0", "requests"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    python_requires=">=3.7",
)
